from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagBold(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L781-L786

    @property
    def tag(self) -> str:
        return "b"


@dataclass
class AssTagItalic(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L787-L792

    @property
    def tag(self) -> str:
        return "i"


@dataclass
class AssTagUnderline(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L838-L845

    @property
    def tag(self) -> str:
        return "u"


@dataclass
class AssTagStrikeout(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L830-L837

    @property
    def tag(self) -> str:
        return "s"


@dataclass
class AssTagFontName:
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L512-L522

    @property
    def tag(self) -> str:
        return "fn"


@dataclass
class AssTagFontEncoding(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L858-L864
    # https://learn.microsoft.com/en-us/openspecs/office_file_formats/ms-one/64e2db6e-6eeb-443c-9ccf-0f72b37ba411

    @property
    def tag(self) -> str:
        return "fe"


@dataclass
class AssTagFontSize(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L421-L432

    @property
    def tag(self) -> str:
        return "fs"


@dataclass
class AssTagLetterSpacing(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L413-L420

    @property
    def tag(self) -> str:
        return "fsp"


@dataclass
class AssTagResetStyle(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L761-L767

    @property
    def tag(self) -> str:
        return "r"


@dataclass
class AssTagAnimation(AssTag):
    # https://github.com/libass/libass/blob/5f57443f1784434fe8961275da08be6d6febc688/libass/ass_parse.c#L626-L684
    @property
    def tag(self) -> str:
        return "t"


@dataclass
class AssTagBaselineOffset(AssTag):
    # https://github.com/libass/libass/blob/5f57443f1784434fe8961275da08be6d6febc688/libass/ass_parse.c#L846-L848

    @property
    def tag(self) -> str:
        return "pbo"


@dataclass
class AssTagRotationOrigin(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L613-L625

    @property
    def tag(self) -> str:
        return "org"


@dataclass
class AssTagDraw(AssTag):
    # https://github.com/libass/libass/blob/5f57443f1784434fe8961275da08be6d6febc688/libass/ass_parse.c#L849-L852

    @property
    def tag(self) -> str:
        return "p"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagFontScale(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L410-L412

    @property
    def tag(self) -> str:
        return "fsc"


@dataclass
class AssTagFontXScale(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L392-L400

    @property
    def tag(self) -> str:
        return "fscx"


@dataclass
class AssTagFontYScale(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L401-L409

    @property
    def tag(self) -> str:
        return "fscy"



from .ass_valid_tag_alignment import AssValidTagAlignment
from .ass_valid_tag_alpha import (
    AssValidTagAlpha,
    AssValidTagPrimaryAlpha,
    AssValidTagSecondaryAlpha,
    AssValidTagOutlineAlpha,
    AssValidTagBackgroundAlpha,
)
from .ass_valid_tag_blur import AssValidTagBlurEdges, AssValidTagBlurEdgesGauss
from .ass_valid_tag_border import (
    AssValidTagBorder,
    AssValidTagXBorder,
    AssValidTagYBorder,
)
from .ass_valid_tag_clip import AssValidTagClipRectangle, AssValidTagClipVector
from .ass_valid_tag_color import (
    AssValidTagPrimaryColor,
    AssValidTagSecondaryColor,
    AssValidTagOutlineColor,
    AssValidTagBackgroundColor,
)
from .ass_valid_tag_fade import AssValidTagFade, AssValidTagFadeComplex
from .ass_valid_tag_font_scale import (
    AssValidTagFontScale,
    AssValidTagFontXScale,
    AssValidTagFontYScale,
)
from .ass_valid_tag_general import (
    AssValidTagAnimation,
    AssValidTagBaselineOffset,
    AssValidTagBold,
    AssValidTagDraw,
    AssValidTagFontEncoding,
    AssValidTagFontName,
    AssValidTagFontSize,
    AssValidTagItalic,
    AssValidTagLetterSpacing,
    AssValidTagResetStyle,
    AssValidTagRotationOrigin,
    AssValidTagStrikeout,
    AssValidTagUnderline,
)
from .ass_valid_tag_karaoke import (
    AssValidTagKaraoke,
    AssValidTagKaraokeFill,
    AssValidTagKaraokeOutline,
)
from .ass_valid_tag_position import AssValidTagMove, AssValidTagPosition
from .ass_valid_tag_rotation import (
    AssValidTagXRotation,
    AssValidTagYRotation,
    AssValidTagZRotation,
)
from .ass_valid_tag_shadow import (
    AssValidTagShadow,
    AssValidTagXShadow,
    AssValidTagYShadow,
)
from .ass_valid_tag_shear import AssValidTagXShear, AssValidTagYShear
from .ass_valid_tag_wrap_style import AssValidTagWrapStyle



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagKaraoke(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L809-L816

    @property
    def tag(self) -> str:
        return "k"


@dataclass
class AssTagKaraokeFill(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L793-L800
    is_short_tag: bool

    @property
    def tag(self) -> str:
        return "kf"

    @property
    def short_tag(self) -> str:
        return "K"


@dataclass
class AssTagKaraokeOutline(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L801-L808

    @property
    def tag(self) -> str:
        return "ko"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagClipRectangle(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L685-L704
    invert: bool

    @property
    def tag(self) -> str:
        if self.invert:
            return "iclip"
        return "clip"


@dataclass
class AssTagClipVector(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L685-L704
    invert: bool

    @property
    def tag(self) -> str:
        if self.invert:
            return "iclip"
        return "clip"



from typing import List

import pytest
import ass_tag_analyzer


@pytest.mark.parametrize(
    "text,expected_result",
    [
        (
            "\\{\\}",
            [
                ass_tag_analyzer.AssText("{}"),
            ],
        ),
        (
            "\\{}",
            [
                ass_tag_analyzer.AssText("{}"),
            ],
        ),
        (
            "\\{\\\\}",
            [
                ass_tag_analyzer.AssText("{\\}"),
            ],
        ),
        (
            "{\\}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test"),
            ],
        ),
        (
            "{\\an5\\pos(960, 96)\\yshad0.01\\4c&HB2FAD9&\\1a&HFE&\\3a&HFE&\\blur2\\fad(200,300)\\t(0,1,\\xshad-1.59)}Test with multiple tag",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAlignment(ass_tag_analyzer.Alignment(5)),
                ass_tag_analyzer.AssValidTagPosition(960, 96),
                ass_tag_analyzer.AssValidTagYShadow(0.01),
                ass_tag_analyzer.AssValidTagBackgroundColor(217, 250, 178),
                ass_tag_analyzer.AssValidTagPrimaryAlpha(254),
                ass_tag_analyzer.AssValidTagOutlineAlpha(254),
                ass_tag_analyzer.AssValidTagBlurEdgesGauss(2),
                ass_tag_analyzer.AssValidTagFade(200, 300),
                ass_tag_analyzer.AssValidTagAnimation(
                    [ass_tag_analyzer.AssValidTagXShadow(-1.59)], time1=0, time2=1
                ),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("Test with multiple tag"),
            ],
        ),
        (
            "text1{\\blur2}text2{\\blur}text3",
            [
                ass_tag_analyzer.AssText("text1"),
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdgesGauss(2),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("text2"),
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdgesGauss(0),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("text3"),
            ],
        ),
        ("{test", [ass_tag_analyzer.AssText("{test")]),
        ("test}", [ass_tag_analyzer.AssText("test}")]),
        (
            "text{\\p1}draw{\\p0}text",
            [
                ass_tag_analyzer.AssText("text"),
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagDraw(1),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssDraw("draw"),
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagDraw(0),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("text"),
            ],
        ),
        (
            "{\\an5}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAlignment(ass_tag_analyzer.Alignment(5)),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test"),
            ],
        ),
        (
            "{\\an}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagAlignment("", False),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test"),
            ],
        ),
        (
            "{\\an0}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagAlignment("0", False),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test"),
            ],
        ),
        (
            "{\\an10}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagAlignment("10", False),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test"),
            ],
        ),
        (
            "{\\anTest}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagAlignment("Test", False),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test"),
            ],
        ),
        (
            "{\\t(\\be20)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAnimation(
                    [ass_tag_analyzer.AssValidTagBlurEdges(20.0)]
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\t(4,\\be20)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAnimation(
                    [ass_tag_analyzer.AssValidTagBlurEdges(20.0)], acceleration=4.0
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\t(0,1000,\\be20)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAnimation(
                    [ass_tag_analyzer.AssValidTagBlurEdges(20.0)], time1=0, time2=1000
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\t(0,1000,4,\\be20)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAnimation(
                    [ass_tag_analyzer.AssValidTagBlurEdges(20.0)],
                    acceleration=4.0,
                    time1=0,
                    time2=1000,
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\t}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAnimation([]),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\t(0,1000,\\be20,20,20,202,202,20,20)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAnimation(
                    [ass_tag_analyzer.AssValidTagBlurEdges(20.0)], time1=0, time2=1000
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\alpha&HB4&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\alpha&HB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\alphaH&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\alpha&hB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAlpha(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\alpha&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\alpha}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\alpha  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1a&HB4&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1a&HB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1aH&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1a&hB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryAlpha(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1a&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1a}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagPrimaryAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1a  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagPrimaryAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2a&HB4&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2a&HB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2aH&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2a&hB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryAlpha(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2a&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2a}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagSecondaryAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2a  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagSecondaryAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3a&HB4&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3a&HB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3aH&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3a&hB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineAlpha(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3a&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3a}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagOutlineAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3a  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagOutlineAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4a&HB4&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4a&HB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4aH&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4a&hB4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundAlpha(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4a&B4}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundAlpha(180),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4a}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagBackgroundAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4a  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagBackgroundAlpha(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1c&H6465D9&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(False, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1c&H6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(False, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1cH&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(False, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1c&h6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(False, 0, 0, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1c&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(False, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1c}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagPrimaryColor("", False),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\1c  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagPrimaryColor("", False),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\c&H6465D9&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(True, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\c&H6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(True, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\cH&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(True, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\c&h6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(True, 0, 0, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\c&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPrimaryColor(True, 217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\c}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagPrimaryColor("", True),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\c  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagPrimaryColor("", True),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2c&H6465D9&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2c&H6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2cH&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2c&h6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryColor(0, 0, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2c&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagSecondaryColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2c}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagSecondaryColor(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\2c  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagSecondaryColor(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3c&H6465D9&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3c&H6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3cH&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3c&h6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineColor(0, 0, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3c&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagOutlineColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3c}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagOutlineColor(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\3c  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagOutlineColor(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4c&H6465D9&}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4c&H6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4cH&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4c&h6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundColor(0, 0, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4c&6465D9}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBackgroundColor(217, 101, 100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4c}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagBackgroundColor(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\4c  }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagBackgroundColor(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\pbo4.5}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBaselineOffset(4.5),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\pbo }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBaselineOffset(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\be }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdges(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\beTest}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdges(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\beTest}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdges(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\be2.2}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdges(2.2),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\blur }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdgesGauss(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\blurTest}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdgesGauss(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\blur2.2}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBlurEdgesGauss(2.2),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\b0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBold(400),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\b1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBold(700),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\b100}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBold(100),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\b99}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagBold("99"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\b }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagBold(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\bord400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagBorder(400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\bord }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagBorder(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\xbord400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagXBorder(400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\xbord }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagXBorder(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\ybord400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagYBorder(400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\ybord }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagYBorder(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\clip(100,200,300,400)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagClipRectangle(False, 100, 200, 300, 400),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\clip  (100,200,300,400)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagClipRectangle(False, 100, 200, 300, 400),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\clip(100,200,300,400}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagClipRectangle(False, 100, 200, 300, 400),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\iclip(100,200,300,400)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagClipRectangle(True, 100, 200, 300, 400),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\clip100,200,300,400)}",
            [ass_tag_analyzer.AssTagListOpening(), ass_tag_analyzer.AssTagListEnding()],
        ),
        (
            "{\\clip(m 187 365 l 215 341 249 301)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagClipVector(
                    False, "m 187 365 l 215 341 249 301"
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\iclip(m 187 365 l 215 341 249 301)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagClipVector(
                    True, "m 187 365 l 215 341 249 301"
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\iclip(20, m 187 365 l 215 341 249 301)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagClipVector(
                    True, "m 187 365 l 215 341 249 301", 20
                ),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\p1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagDraw(1),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\p}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagDraw(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fad(0, 500)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFade(0, 500),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fad(0, 500}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFade(0, 500),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fad0, 500}",
            [ass_tag_analyzer.AssTagListOpening(), ass_tag_analyzer.AssTagListEnding()],
        ),
        (
            "{\\fade(255,0,255,0,500,1000,1500)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFadeComplex(255, 0, 255, 0, 500, 1000, 1500),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fade(255,0,255,0,500,1000,1500}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFadeComplex(255, 0, 255, 0, 500, 1000, 1500),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fade255,0,255,0,500,1000,1500)}",
            [ass_tag_analyzer.AssTagListOpening(), ass_tag_analyzer.AssTagListEnding()],
        ),
        (
            "{\\fe0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontEncoding(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fe k}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontEncoding(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fe }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagFontEncoding(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fnArial}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontName("Arial"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fn Arial }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontName("Arial"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fn 0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontName("0"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fn0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagFontName("0"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fn }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagFontName(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fscExample}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontScale(),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fsc}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontScale(),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fs20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontSize(20.1),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fs j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontSize(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fs }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagFontSize(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fscx20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontXScale(20.1),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fscx j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontXScale(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fscx }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagFontXScale(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fscy20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontYScale(20.1),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fscy j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagFontYScale(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fscy }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagFontYScale(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\i1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagItalic(True),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\i0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagItalic(False),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\i2}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagItalic("2"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\i }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagItalic(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\k20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraoke(201),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\k j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraoke(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\k }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraoke(1000),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\kf20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeFill(False, 201),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\kf j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeFill(False, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\kf }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeFill(False, 1000),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\K20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeFill(True, 201),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\K j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeFill(True, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\K }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeFill(True, 1000),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\ko20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeOutline(201),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\ko j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeOutline(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\ko }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagKaraokeOutline(1000),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\move(0.1, 5, 10, 20)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagMove(0.1, 5, 10, 20),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\move(0.1, 5, 10, 20, 30, 0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagMove(0.1, 5, 10, 20, 30, 0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\move0, 500}",
            [ass_tag_analyzer.AssTagListOpening(), ass_tag_analyzer.AssTagListEnding()],
        ),
        (
            "{\\pos(0.1, 5)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPosition(0.1, 5),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\pos(0.1, 5}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagPosition(0.1, 5),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\pos0, 500}",
            [ass_tag_analyzer.AssTagListOpening(), ass_tag_analyzer.AssTagListEnding()],
        ),
        (
            "{\\rDefault}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagResetStyle("Default"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\r Arial }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagResetStyle(" Arial"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\r }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagResetStyle(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\org(0.1, 5)}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagRotationOrigin(0.1, 5),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\org(0.1, 5, )}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagRotationOrigin(0.1, 5),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\org(0.1, 5}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagRotationOrigin(0.1, 5),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\org0, 500}",
            [ass_tag_analyzer.AssTagListOpening(), ass_tag_analyzer.AssTagListEnding()],
        ),
        (
            "{\\org(0.1, 5, 20)}",
            [ass_tag_analyzer.AssTagListOpening(), ass_tag_analyzer.AssTagListEnding()],
        ),
        (
            "{\\shad20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagShadow(20.1),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\shad j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagShadow(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\shad }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagShadow(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\xshad20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagXShadow(20.1),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\xshad j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagXShadow(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\xshad }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagXShadow(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\yshad20.1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagYShadow(20.1),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\yshad j}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagYShadow(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\yshad }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagYShadow(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\s1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagStrikeout(True),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\s0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagStrikeout(False),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\s2}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagStrikeout("2"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\s }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagStrikeout(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\u1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagUnderline(True),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\u0}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagUnderline(False),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\u2}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagUnderline("2"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\u }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagUnderline(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\q1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagWrapStyle(ass_tag_analyzer.WrapStyle(1)),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\q-1}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagWrapStyle("-1"),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\q }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagWrapStyle(""),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\frx400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagXRotation(400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\frx }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagXRotation(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fry400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagYRotation(400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fry }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagYRotation(0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fr400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagZRotation(True, 400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fr }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagZRotation("", True),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\frz400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagZRotation(False, 400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\frz }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssInvalidTagZRotation("", False),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fax400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagXShear(400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fax }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagXShear(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fay400.7}",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagYShear(400.7),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\fay }",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagYShear(0.0),
                ass_tag_analyzer.AssTagListEnding(),
            ],
        ),
        (
            "{\\}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test"),
            ],
        ),
        (
            "{this is a comment}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssComment("this is a comment"),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test")
            ],
        ),
        (
            "{ comment1 \\t(20,20,20,230,3.4.3,\\fscx10) comment2 \\t(20,20,20,230,3.4.3,\\fscx10) comment3 }test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssComment(" comment1 "),
                ass_tag_analyzer.AssValidTagAnimation([]),
                ass_tag_analyzer.AssComment(" comment2 "),
                ass_tag_analyzer.AssValidTagAnimation([]),
                ass_tag_analyzer.AssComment(" comment3 "),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test")
            ],
        ),
        (
            "{\\t(20,20,20,230,3.4.3,\\fscx10))}test",
            [
                ass_tag_analyzer.AssTagListOpening(),
                ass_tag_analyzer.AssValidTagAnimation([]),
                ass_tag_analyzer.AssComment(")"),
                ass_tag_analyzer.AssTagListEnding(),
                ass_tag_analyzer.AssText("test")
            ],
        ),
    ],
)
def test_parse_tag(text: str, expected_result: List[ass_tag_analyzer.AssItem]):
    assert ass_tag_analyzer.parse_line(text) == expected_result



from typing import List, Tuple

from ass_tag_analyzer.ass_item import (
    AssTagAlignment,
)
from ass_tag_analyzer.ass_item.ass_tag_alignment import Alignment, LegacyAlignment
from ass_tag_analyzer.ass_item.ass_tag_general import AssTagRotationOrigin
from ass_tag_analyzer.ass_item.ass_tag_position import AssTagMove, AssTagPosition
from ass_tag_analyzer.ass_item.ass_tag_wrap_style import WrapStyle
from ass_tag_analyzer.ass_item.ass_valid_tag import (
    AssValidTagAlignment,
    AssValidTagAlpha,
    AssValidTagAnimation,
    AssValidTagBackgroundAlpha,
    AssValidTagBackgroundColor,
    AssValidTagBaselineOffset,
    AssValidTagBlurEdges,
    AssValidTagBlurEdgesGauss,
    AssValidTagBold,
    AssValidTagBorder,
    AssValidTagClipRectangle,
    AssValidTagClipVector,
    AssValidTagDraw,
    AssValidTagFade,
    AssValidTagFadeComplex,
    AssValidTagFontEncoding,
    AssValidTagFontName,
    AssValidTagFontScale,
    AssValidTagFontSize,
    AssValidTagFontXScale,
    AssValidTagFontYScale,
    AssValidTagItalic,
    AssValidTagKaraoke,
    AssValidTagKaraokeFill,
    AssValidTagKaraokeOutline,
    AssValidTagLetterSpacing,
    AssValidTagMove,
    AssValidTagOutlineAlpha,
    AssValidTagOutlineColor,
    AssValidTagPosition,
    AssValidTagPrimaryAlpha,
    AssValidTagPrimaryColor,
    AssValidTagResetStyle,
    AssValidTagRotationOrigin,
    AssValidTagSecondaryAlpha,
    AssValidTagSecondaryColor,
    AssValidTagShadow,
    AssValidTagStrikeout,
    AssValidTagUnderline,
    AssValidTagWrapStyle,
    AssValidTagXBorder,
    AssValidTagXRotation,
    AssValidTagXShadow,
    AssValidTagXShear,
    AssValidTagYBorder,
    AssValidTagYRotation,
    AssValidTagYShadow,
    AssValidTagYShear,
    AssValidTagZRotation,
)
from ass_tag_analyzer.ass_item.ass_invalid_tag import (
    AssInvalidTagAlignment,
    AssInvalidTagAlpha,
    AssInvalidTagBackgroundAlpha,
    AssInvalidTagBackgroundColor,
    AssInvalidTagBold,
    AssInvalidTagBorder,
    AssInvalidTagFontEncoding,
    AssInvalidTagFontName,
    AssInvalidTagFontSize,
    AssInvalidTagFontXScale,
    AssInvalidTagFontYScale,
    AssInvalidTagItalic,
    AssInvalidTagLetterSpacing,
    AssInvalidTagOutlineAlpha,
    AssInvalidTagOutlineColor,
    AssInvalidTagPrimaryAlpha,
    AssInvalidTagPrimaryColor,
    AssInvalidTagResetStyle,
    AssInvalidTagSecondaryAlpha,
    AssInvalidTagSecondaryColor,
    AssInvalidTagShadow,
    AssInvalidTagStrikeout,
    AssInvalidTagUnderline,
    AssInvalidTagWrapStyle,
    AssInvalidTagXBorder,
    AssInvalidTagXShadow,
    AssInvalidTagYBorder,
    AssInvalidTagYShadow,
    AssInvalidTagZRotation,
)
from ass_tag_analyzer.ass_item.ass_item import (
    AssComment,
    AssDraw,
    AssItem,
    AssTag,
    AssTagListEnding,
    AssTagListOpening,
    AssText,
)


from ass_tag_analyzer.ass_type_parser import TypeParser


def parse_tags(text: str, is_draw: bool) -> Tuple[List[AssTag], bool]:
    """
    Inpired by:
        - https://github.com/libass/libass/blob/5f57443f1784434fe8961275da08be6d6febc688/libass/ass_parse.c#L242-L869
        - https://sourceforge.net/p/guliverkli2/code/HEAD/tree/src/subtitles/RTS.cpp#l1383


    Parameters:
        text (str): An string of tags.
            Ex: "\\blur1\\pos(604, 20)"
    Returns:
        An tuple of two element
            1- An list of AssTag which represent the text
            2- An and bool. If true, then the text right after the tags will be an AssText, if false, then the text right after the tags will be an AssDraw
    """

    tags: List[AssTag] = []
    i = 0

    while (j := text.find("\\", i)) >= 0:
        if i != j:
            tags.append(AssComment(text[i:j]))

        # Skip the \
        j += 1
        k = j
        while k < len(text) and text[k] not in ("(", "\\"):
            k += 1
        cmd = TypeParser.strip_whitespace(text[j:k])
        j = k

        if len(cmd) == 0:
            i = j
            continue

        params: List[str] = []

        if j < len(text) and text[j] == "(":
            # Skip the (
            j += 1
            k = j
            while k < len(text) and text[k] != ")":
                k += 1
            param = text[j:k]
            param = TypeParser.strip_whitespace(text[j:k])

            if k < len(text) and text[k] == ")":
                # VSFilter doesn't do this, but it is required to parse the comments properly
                k += 1
            j = k

            temp_j = j

            while len(param) != 0:
                i = param.find(",")
                j = param.find("\\")

                if i >= 0 and (j < 0 or i < j):
                    s = TypeParser.strip_whitespace(param[:i])
                    if len(s) != 0:
                        params.append(s)
                    param = param[i + 1 :] if i + 1 < len(param) else ""
                else:
                    param = TypeParser.strip_whitespace(param)
                    if len(param) != 0:
                        params.append(param)
                    param = ""

            j = temp_j

        if cmd.startswith(("1c", "2c", "3c", "4c")):
            params.append(cmd[2:].strip("&H"))
            cmd = cmd[:2]
        elif cmd.startswith(("1a", "2a", "3a", "4a")):
            params.append(cmd[2:].strip("&H"))
            cmd = cmd[:2]
        elif cmd.startswith("alpha"):
            params.append(cmd[5:].strip("&H"))
            cmd = cmd[:5]
        elif cmd.startswith("an"):
            params.append(cmd[2:])
            cmd = cmd[:2]
        elif cmd.startswith("a"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith("blur"):
            params.append(cmd[4:])
            cmd = cmd[:4]
        elif cmd.startswith("bord"):
            params.append(cmd[4:])
            cmd = cmd[:4]
        elif cmd.startswith("be"):
            params.append(cmd[2:])
            cmd = cmd[:2]
        elif cmd.startswith("b"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith("clip"):
            pass
        elif cmd.startswith("c"):
            params.append(cmd[1:].strip("&H"))
            cmd = cmd[:1]
        elif cmd.startswith("fade"):
            pass
        elif cmd.startswith("fe"):
            params.append(cmd[2:])
            cmd = cmd[:2]
        elif cmd.startswith("fn"):
            params.append(cmd[2:])
            cmd = cmd[:2]
        elif cmd.startswith(("frx", "fry", "frz")):
            params.append(cmd[3:])
            cmd = cmd[:3]
        elif cmd.startswith(("fax", "fay")):
            params.append(cmd[3:])
            cmd = cmd[:3]
        elif cmd.startswith("fr"):
            params.append(cmd[2:])
            cmd = cmd[:2]
        elif cmd.startswith(("fscx", "fscy")):
            params.append(cmd[4:])
            cmd = cmd[:4]
        elif cmd.startswith("fsc"):
            params.append(cmd[3:])
            cmd = cmd[:3]
        elif cmd.startswith("fsp"):
            params.append(cmd[3:])
            cmd = cmd[:3]
        elif cmd.startswith("fs"):
            params.append(cmd[2:])
            cmd = cmd[:2]
        elif cmd.startswith("iclip"):
            pass
        elif cmd.startswith("i"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith(("kt", "kf", "ko")):
            params.append(cmd[2:])
            cmd = cmd[:2]
        elif cmd.startswith(("k", "K")):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith("move"):
            pass
        elif cmd.startswith("org"):
            pass
        elif cmd.startswith("pbo"):
            params.append(cmd[3:])
            cmd = cmd[:3]
        elif cmd.startswith("pos"):
            pass
        elif cmd.startswith("p"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith("q"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith("r"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith("shad"):
            params.append(cmd[4:])
            cmd = cmd[:4]
        elif cmd.startswith("s"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith("t"):
            pass
        elif cmd.startswith("u"):
            params.append(cmd[1:])
            cmd = cmd[:1]
        elif cmd.startswith(("xbord", "xshad", "ybord", "yshad")):
            params.append(cmd[5:])
            cmd = cmd[:5]

        p = params[0] if len(params) > 0 else ""

        if cmd in ("1c", "2c", "3c", "4c"):
            i = ord(cmd[0]) - ord("0")

            if len(p) == 0:
                if i == 1:
                    tags.append(AssInvalidTagPrimaryColor(p, False))
                elif i == 2:
                    tags.append(AssInvalidTagSecondaryColor(p))
                elif i == 3:
                    tags.append(AssInvalidTagOutlineColor(p))
                elif i == 4:
                    tags.append(AssInvalidTagBackgroundColor(p))
            else:

                red, green, blue = TypeParser.color_arg(TypeParser.hex_str_to_int(p))

                if i == 1:
                    tags.append(AssValidTagPrimaryColor(False, red, green, blue))
                elif i == 2:
                    tags.append(AssValidTagSecondaryColor(red, green, blue))
                elif i == 3:
                    tags.append(AssValidTagOutlineColor(red, green, blue))
                elif i == 4:
                    tags.append(AssValidTagBackgroundColor(red, green, blue))

        elif cmd in ("1a", "2a", "3a", "4a"):
            i = ord(cmd[0]) - ord("0")

            if len(p) == 0:
                if i == 1:
                    tags.append(AssInvalidTagPrimaryAlpha(p))
                elif i == 2:
                    tags.append(AssInvalidTagSecondaryAlpha(p))
                elif i == 3:
                    tags.append(AssInvalidTagOutlineAlpha(p))
                elif i == 4:
                    tags.append(AssInvalidTagBackgroundAlpha(p))
            else:

                alpha = TypeParser.hex_str_to_int(p)

                if i == 1:
                    tags.append(AssValidTagPrimaryAlpha(alpha))
                elif i == 2:
                    tags.append(AssValidTagSecondaryAlpha(alpha))
                elif i == 3:
                    tags.append(AssValidTagOutlineAlpha(alpha))
                elif i == 4:
                    tags.append(AssValidTagBackgroundAlpha(alpha))

        elif cmd == "alpha":
            if len(p) == 0:
                tags.append(AssInvalidTagAlpha(p))
            else:
                tags.append(AssValidTagAlpha(TypeParser.hex_str_to_int(p)))
        elif cmd == "an":
            try:
                an_tag = AssValidTagAlignment(Alignment(TypeParser.int_str_to_int(p)))
            except:
                an_tag = AssInvalidTagAlignment(p, False)

            if not any(isinstance(tag, AssTagAlignment) for tag in tags):
                tags.append(an_tag)

        elif cmd == "a":
            try:
                an_tag = AssValidTagAlignment(
                    LegacyAlignment(TypeParser.int_str_to_int(p)), True
                )
            except:
                an_tag = AssInvalidTagAlignment(p, True)

            if not any(isinstance(tag, AssTagAlignment) for tag in tags):
                tags.append(an_tag)

        elif cmd == "blur":
            if len(p) == 0:
                tags.append(AssValidTagBlurEdgesGauss(0.0))
            else:
                tags.append(AssValidTagBlurEdgesGauss(TypeParser.float_str_to_float(p)))

        elif cmd == "bord":
            if len(p) == 0:
                tags.append(AssInvalidTagBorder(p))
            else:
                tags.append(AssValidTagBorder(TypeParser.float_str_to_float(p)))

        elif cmd == "be":
            if len(p) == 0:
                tags.append(AssValidTagBlurEdges(0.0))
            else:
                tags.append(AssValidTagBlurEdges(TypeParser.float_str_to_float(p)))

        elif cmd == "b":
            if len(p) == 0:
                tags.append(AssInvalidTagBold(p))
            else:
                try:
                    tags.append(AssValidTagBold(TypeParser.int_str_to_int(p)))
                except:
                    tags.append(AssInvalidTagBold(p))

        elif cmd in ("clip", "iclip"):

            invert = cmd == "iclip"

            if len(params) == 1:
                tags.append(AssValidTagClipVector(invert, params[0]))
            elif len(params) == 2:
                scale = max(TypeParser.int_str_to_int(p), 1)
                tags.append(AssValidTagClipVector(invert, params[1], scale))
            elif len(params) == 4:
                x1 = TypeParser.int_str_to_int(params[0])
                y1 = TypeParser.int_str_to_int(params[1])
                x2 = TypeParser.int_str_to_int(params[2])
                y2 = TypeParser.int_str_to_int(params[3])

                tags.append(AssValidTagClipRectangle(invert, x1, y1, x2, y2))
        elif cmd == "c":
            if len(p) == 0:
                tags.append(AssInvalidTagPrimaryColor(p, True))
            else:
                red, green, blue = TypeParser.color_arg((TypeParser.hex_str_to_int(p)))
                tags.append(AssValidTagPrimaryColor(True, red, green, blue))

        elif cmd in ("fade", "fad"):
            # // {\fade(a1=param[0], a2=param[1], a3=param[2], t1=t[0], t2=t[1], t3=t[2], t4=t[3])
            if len(params) == 7:
                a1 = TypeParser.int_str_to_int(params[0])
                a2 = TypeParser.int_str_to_int(params[1])
                a3 = TypeParser.int_str_to_int(params[2])
                t1 = TypeParser.int_str_to_int(params[3])
                t2 = TypeParser.int_str_to_int(params[4])
                t3 = TypeParser.int_str_to_int(params[5])
                t4 = TypeParser.int_str_to_int(params[6])
                tags.append(AssValidTagFadeComplex(a1, a2, a3, t1, t2, t3, t4))
            elif len(params) == 2:
                t1 = TypeParser.int_str_to_int(params[0])
                t2 = TypeParser.int_str_to_int(params[1])
                tags.append(AssValidTagFade(t1, t2))
        elif cmd == "fax":
            if len(p) == 0:
                tags.append(AssValidTagXShear(0.0))
            else:
                tags.append(AssValidTagXShear(TypeParser.float_str_to_float(p)))
        elif cmd == "fay":
            if len(p) == 0:
                tags.append(AssValidTagYShear(0.0))
            else:
                tags.append(AssValidTagYShear(TypeParser.float_str_to_float(p)))
        elif cmd == "fe":
            if len(p) == 0:
                tags.append(AssInvalidTagFontEncoding(p))
            else:
                tags.append(AssValidTagFontEncoding(TypeParser.int_str_to_int(p)))

        elif cmd == "fn":
            # https://sourceforge.net/p/guliverkli2/code/HEAD/tree/src/subtitles/RTS.cpp#l1683
            if len(p) == 0 or p == "0":
                tags.append(AssInvalidTagFontName(p))
            else:
                tags.append(AssValidTagFontName(p))

        elif cmd == "frx":
            if len(p) == 0:
                tags.append(AssValidTagXRotation(0.0))
            else:
                tags.append(AssValidTagXRotation(TypeParser.float_str_to_float(p)))
        elif cmd == "fry":
            if len(p) == 0:
                tags.append(AssValidTagYRotation(0.0))
            else:
                tags.append(AssValidTagYRotation(TypeParser.float_str_to_float(p)))
        elif cmd == "frz":
            if len(p) == 0:
                tags.append(AssInvalidTagZRotation(p, False))
            else:
                tags.append(
                    AssValidTagZRotation(False, TypeParser.float_str_to_float(p))
                )
        elif cmd == "fr":
            if len(p) == 0:
                tags.append(AssInvalidTagZRotation(p, True))
            else:
                tags.append(
                    AssValidTagZRotation(True, TypeParser.float_str_to_float(p))
                )

        elif cmd == "fscx":
            if len(p) == 0:
                tags.append(AssInvalidTagFontXScale(p))
            else:
                tags.append(AssValidTagFontXScale(TypeParser.float_str_to_float(p)))

        elif cmd == "fscy":
            if len(p) == 0:
                tags.append(AssInvalidTagFontYScale(p))
            else:
                tags.append(AssValidTagFontYScale(TypeParser.float_str_to_float(p)))
        elif cmd == "fsc":
            tags.append(AssValidTagFontScale())
        elif cmd == "fsp":
            if len(p) == 0:
                tags.append(AssInvalidTagLetterSpacing(p))
            else:
                tags.append(AssValidTagLetterSpacing(TypeParser.float_str_to_float(p)))
        elif cmd == "fs":
            if len(p) == 0:
                tags.append(AssInvalidTagFontSize(p))
            else:
                tags.append(AssValidTagFontSize(TypeParser.float_str_to_float(p)))
        elif cmd == "i":
            n = TypeParser.int_str_to_int(p)

            if len(p) == 0 or n not in (0, 1):
                tags.append(AssInvalidTagItalic(p))
            else:
                tags.append(AssValidTagItalic(bool(n)))

        elif cmd in ("kf", "K"):
            short_tag = cmd == "K"
            n = TypeParser.float_str_to_float(p)

            if len(p) == 0:
                n = 100

            tags.append(AssValidTagKaraokeFill(short_tag, n * 10))

        elif cmd == "ko":
            n = TypeParser.float_str_to_float(p)

            if len(p) == 0:
                n = 100

            tags.append(AssValidTagKaraokeOutline(n * 10))

        elif cmd == "k":
            n = TypeParser.float_str_to_float(p)

            if len(p) == 0:
                n = 100

            tags.append(AssValidTagKaraoke(n * 10))

        elif cmd == "move":
            if len(params) in (4, 6) and not any(
                isinstance(tag, AssTagMove) for tag in tags
            ):
                x1 = TypeParser.float_str_to_float(params[0])
                y1 = TypeParser.float_str_to_float(params[1])
                x2 = TypeParser.float_str_to_float(params[2])
                y2 = TypeParser.float_str_to_float(params[3])
                t1 = t2 = None
                if len(params) == 6:
                    t1 = TypeParser.int_str_to_int(params[4])
                    t2 = TypeParser.int_str_to_int(params[5])
                tags.append(AssValidTagMove(x1, y1, x2, y2, t1, t2))

        elif cmd == "org":
            if len(params) == 2 and not any(
                isinstance(tag, AssTagRotationOrigin) for tag in tags
            ):
                x = TypeParser.float_str_to_float(params[0])
                y = TypeParser.float_str_to_float(params[1])

                tags.append(AssValidTagRotationOrigin(x, y))

        elif cmd == "pbo":
            tags.append(AssValidTagBaselineOffset(TypeParser.float_str_to_float(p)))
        elif cmd == "pos":
            if len(params) == 2 and not any(
                isinstance(tag, AssTagPosition) for tag in tags
            ):
                x = TypeParser.float_str_to_float(params[0])
                y = TypeParser.float_str_to_float(params[1])

                tags.append(AssValidTagPosition(x, y))

        elif cmd == "p":
            draw_tag = AssValidTagDraw(TypeParser.int_str_to_int(p))
            tags.append(draw_tag)

            is_draw = draw_tag.scale != 0

        elif cmd == "q":
            if len(p) == 0:
                tags.append(AssInvalidTagWrapStyle(p))
            else:
                try:
                    tags.append(
                        AssValidTagWrapStyle(WrapStyle(TypeParser.int_str_to_int(p)))
                    )
                except:
                    tags.append(AssInvalidTagWrapStyle(p))

        elif cmd == "r":
            if len(p) == 0:
                tags.append(AssInvalidTagResetStyle(p))
            else:
                tags.append(AssValidTagResetStyle(p))

        elif cmd == "shad":
            if len(p) == 0:
                tags.append(AssInvalidTagShadow(p))
            else:
                tags.append(AssValidTagShadow(TypeParser.float_str_to_float(p)))

        elif cmd == "s":
            n = TypeParser.int_str_to_int(p)

            if len(p) == 0 or n not in (0, 1):
                tags.append(AssInvalidTagStrikeout(p))
            else:
                tags.append(AssValidTagStrikeout(bool(n)))

        elif cmd == "t":
            p = ""

            m_animStart = m_animEnd = None
            m_animAccel = 1.0

            if len(params) == 1:
                p = params[0]
            elif len(params) == 2:
                m_animAccel = TypeParser.float_str_to_float(params[0])
                p = params[1]
            elif len(params) == 3:
                m_animStart = TypeParser.int_str_to_int(params[0])
                m_animEnd = TypeParser.int_str_to_int(params[1])
                p = params[2]
            elif len(params) == 4:
                m_animStart = TypeParser.int_str_to_int(params[0])
                m_animEnd = TypeParser.int_str_to_int(params[1])
                m_animAccel = TypeParser.float_str_to_float(params[2])
                p = params[3]

            t_tags, is_draw = parse_tags(p, is_draw)

            tags.append(
                AssValidTagAnimation(t_tags, m_animAccel, m_animStart, m_animEnd)
            )
        elif cmd == "u":
            n = TypeParser.int_str_to_int(p)

            if len(p) == 0 or n not in (0, 1):
                tags.append(AssInvalidTagUnderline(p))
            else:
                tags.append(AssValidTagUnderline(bool(n)))
        elif cmd == "xbord":
            if len(p) == 0:
                tags.append(AssInvalidTagXBorder(p))
            else:
                tags.append(AssValidTagXBorder(TypeParser.float_str_to_float(p)))
        elif cmd == "xshad":
            if len(p) == 0:
                tags.append(AssInvalidTagXShadow(p))
            else:
                tags.append(AssValidTagXShadow(TypeParser.float_str_to_float(p)))

        elif cmd == "ybord":
            if len(p) == 0:
                tags.append(AssInvalidTagYBorder(p))
            else:
                tags.append(AssValidTagYBorder(TypeParser.float_str_to_float(p)))
        elif cmd == "yshad":
            if len(p) == 0:
                tags.append(AssInvalidTagYShadow(p))
            else:
                tags.append(AssValidTagYShadow(TypeParser.float_str_to_float(p)))

        i = j

    if i != len(text):
        tags.append(AssComment(text[i:]))

    return tags, is_draw


def parse_line(text: str) -> List[AssItem]:
    """

    Inpired by:
        - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_render.c#L2044-L2064
        - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L1050-L1086

    Parameters:
        text (str): A .ass line
    Returns:
        An list of AssItem which represent the line
            Ex: "{\\blur1\\pos(604, 20)}Example"
    """

    def get_class_type():
        if is_draw:
            return AssDraw
        return AssText

    ass_items = []
    is_draw = False

    i = 0

    current_text = ""
    while i < len(text):
        if text[i] == "{" and (find_right_bracket_index := text.find("}", i + 1)) != -1:

            if len(current_text) > 0:
                ass_items.append(get_class_type()(current_text))
                current_text = ""


            tag_list_opening = AssTagListOpening()
            ass_items.append(tag_list_opening)

            tags, is_draw = parse_tags(
                text[i + 1 : find_right_bracket_index],
                is_draw,
            )

            ass_items.extend(tags)

            tag_list_ending = AssTagListEnding()
            ass_items.append(tag_list_ending)

            i = find_right_bracket_index + 1
        else:
            if text[i] == "\\" and i + 1 < len(text) and text[i + 1] in ("{", "}"):
                current_text += text[i+1]
                i += 2
            else:
                current_text += text[i]
                i += 1

    if len(current_text) > 0:
        ass_items.append(get_class_type()(current_text))



    return ass_items


def ass_item_to_text(tags: List[AssItem]):
    """
    Parameters:
        tags (List[AssItem]): A list of tags
    Returns:
        An string that represent the AssItem list
    """
    return "".join(str(tag) for tag in tags)



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagMove(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L445-L486

    @property
    def tag(self) -> str:
        return "move"


@dataclass
class AssTagPosition(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L562-L577

    @property
    def tag(self) -> str:
        return "pos"



from ..ass_tag_karaoke import AssTagKaraoke, AssTagKaraokeFill, AssTagKaraokeOutline
from dataclasses import dataclass


@dataclass
class AssValidTagKaraoke(AssTagKaraoke):
    duration: float  # In ms

    def __str__(self):
        return f"\\{self.tag}{self.duration // 10}"


@dataclass
class AssValidTagKaraokeFill(AssTagKaraokeFill):
    duration: float  # In ms

    def __str__(self):
        if self.is_short_tag:
            return f"\\{self.short_tag}{self.duration // 10}"
        return f"\\{self.tag}{self.duration // 10}"


@dataclass
class AssValidTagKaraokeOutline(AssTagKaraokeOutline):
    duration: float  # In ms

    def __str__(self):
        return f"\\{self.tag}{self.duration // 10}"



from ass_tag_analyzer.ass_item.ass_item import AssItem
from ass_tag_analyzer.ass_format import Format
from ..ass_tag_general import (
    AssTagAnimation,
    AssTagBaselineOffset,
    AssTagBold,
    AssTagDraw,
    AssTagFontEncoding,
    AssTagFontName,
    AssTagFontSize,
    AssTagItalic,
    AssTagLetterSpacing,
    AssTagResetStyle,
    AssTagRotationOrigin,
    AssTagStrikeout,
    AssTagUnderline,
)
from dataclasses import dataclass
from typing import List, Optional


from ass_tag_analyzer.ass_type_parser import TypeParser


@dataclass
class AssValidTagBold(AssTagBold):
    __weight: int

    def __init__(self, weight: int):
        self.weight = weight

    @property
    def weight(self):
        return self.__weight

    @weight.setter
    def weight(self, weight: int):
        weight = 400 if weight == 0 else weight
        weight = 700 if weight == 1 else weight

        if weight < 100:
            raise ValueError("Bold tag cannot be under 100 except 0 and 1")

        self.__weight = weight

    def __str__(self):
        weight = self.weight
        weight = 0 if weight == 400 else weight
        weight = 1 if weight == 700 else weight

        return f"\\{self.tag}{weight}"


@dataclass
class AssValidTagItalic(AssTagItalic):
    enabled: bool

    def __str__(self):
        return f"\\{self.tag}{int(self.enabled)}"


@dataclass
class AssValidTagUnderline(AssTagUnderline):
    enabled: bool

    def __str__(self):
        return f"\\{self.tag}{int(self.enabled)}"


@dataclass
class AssValidTagStrikeout(AssTagStrikeout):
    enabled: bool

    def __str__(self):
        return f"\\{self.tag}{int(self.enabled)}"


@dataclass
class AssValidTagFontName(AssTagFontName):
    __name: str

    def __init__(self, name: str):
        self.name = name

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = TypeParser.strip_whitespace(name)

    def __str__(self):
        return f"\\{self.tag}{self.name}"


@dataclass
class AssValidTagFontEncoding(AssTagFontEncoding):
    encoding: int

    def __str__(self):
        return f"\\{self.tag}{self.encoding}"


@dataclass
class AssValidTagFontSize(AssTagFontSize):
    size: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.size)}"


@dataclass
class AssValidTagLetterSpacing(AssTagLetterSpacing):
    spacing: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.spacing)}"


@dataclass
class AssValidTagResetStyle(AssTagResetStyle):
    style: str

    def __str__(self):
        return f"\\{self.tag}{self.style}"


@dataclass
class AssValidTagAnimation(AssTagAnimation):
    tags: List[AssItem]
    acceleration: float = 1.0
    time1: Optional[int] = None
    time2: Optional[int] = None

    def __str__(self):
        tags_text = "".join([str(tag) for tag in self.tags])

        if (
            self.time1 is not None
            and self.time2 is not None
            and self.acceleration is not None
            and self.acceleration != 1
        ):
            return f"\\{self.tag}({self.time1},{self.time2},{Format.format_float(self.acceleration)},{tags_text})"
        elif self.time1 is not None and self.time2 is not None:
            return f"\\{self.tag}({self.time1},{self.time2},{tags_text})"
        elif self.acceleration is not None and self.acceleration != 1:
            return f"\\{self.tag}({Format.format_float(self.acceleration)},{tags_text})"
        else:
            return f"\\{self.tag}({tags_text})"


@dataclass
class AssValidTagBaselineOffset(AssTagBaselineOffset):
    offset: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.offset)}"


@dataclass
class AssValidTagRotationOrigin(AssTagRotationOrigin):
    x: float
    y: float

    def __str__(self):
        return (
            f"\\{self.tag}({Format.format_float(self.x)},{Format.format_float(self.y)})"
        )


@dataclass
class AssValidTagDraw(AssTagDraw):
    __scale: int

    def __init__(self, scale: int):
        self.scale = scale

    @property
    def scale(self):
        return self.__scale

    @scale.setter
    def scale(self, value: int):
        self.__scale = 0 if value < 0 else value

    def __str__(self):
        return f"\\{self.tag}{self.scale}"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagBorder(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L433-L444

    @property
    def tag(self) -> str:
        return "bord"


@dataclass
class AssTagXBorder(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L311-L319

    @property
    def tag(self) -> str:
        return "xbord"


@dataclass
class AssTagYBorder(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L320-L328

    @property
    def tag(self) -> str:
        return "ybord"



from ass_tag_analyzer.ass_format import Format
from ..ass_tag_font_scale import AssTagFontScale, AssTagFontXScale, AssTagFontYScale
from dataclasses import dataclass


@dataclass
class AssValidTagFontScale(AssTagFontScale):
    def __str__(self):
        return f"\\{self.tag}"


@dataclass
class AssValidTagFontXScale(AssTagFontXScale):
    __scale: float

    def __init__(self, scale: float):
        self.scale = scale

    @property
    def scale(self):
        return self.__scale

    @scale.setter
    def scale(self, value: float):
        self.__scale = 0 if value < 0 else value

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.scale)}"


@dataclass
class AssValidTagFontYScale(AssTagFontYScale):
    __scale: float

    def __init__(self, scale: float):
        self.scale = scale

    @property
    def scale(self):
        return self.__scale

    @scale.setter
    def scale(self, value: float):
        self.__scale = 0 if value < 0 else value

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.scale)}"



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_general import (
    AssTagBold,
    AssTagFontEncoding,
    AssTagFontName,
    AssTagFontSize,
    AssTagItalic,
    AssTagLetterSpacing,
    AssTagResetStyle,
    AssTagStrikeout,
    AssTagUnderline,
)
from dataclasses import dataclass


@dataclass
class AssInvalidTagBold(AssTagBold, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagFontEncoding(AssTagFontEncoding, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagFontName(AssTagFontName, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagFontSize(AssTagFontSize, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagItalic(AssTagItalic, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagLetterSpacing(AssTagLetterSpacing, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagResetStyle(AssTagResetStyle, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagStrikeout(AssTagStrikeout, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagUnderline(AssTagUnderline, AssInvalidTag):
    pass



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_alignment import AssTagAlignment
from dataclasses import dataclass


@dataclass
class AssInvalidTagAlignment(AssTagAlignment, AssInvalidTag):
    pass



from ass_tag_analyzer.ass_format import Format
from ..ass_tag_shadow import AssTagShadow, AssTagXShadow, AssTagYShadow
from dataclasses import dataclass


@dataclass
class AssValidTagShadow(AssTagShadow):
    __size: float

    def __init__(self, size: float):
        self.size = size

    @property
    def size(self):
        return self.__size

    @size.setter
    def size(self, value: float):
        self.__size = 0 if value < 0 else value

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.size)}"


@dataclass
class AssValidTagXShadow(AssTagXShadow):
    __size: float

    def __init__(self, size: float):
        self.size = size

    @property
    def size(self):
        return self.__size

    @size.setter
    def size(self, value: float):
        self.__size = 0 if value < 0 else value

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.size)}"


@dataclass
class AssValidTagYShadow(AssTagYShadow):
    __size: float

    def __init__(self, size: float):
        self.size = size

    @property
    def size(self):
        return self.__size

    @size.setter
    def size(self, value: float):
        self.__size = 0 if value < 0 else value

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.size)}"



from ass_tag_analyzer.ass_format import Format
from ..ass_tag_position import AssTagMove, AssTagPosition
from dataclasses import dataclass
from typing import Optional


@dataclass
class AssValidTagMove(AssTagMove):
    x1: float
    y1: float
    x2: float
    y2: float
    __time1: Optional[int] = None
    __time2: Optional[int] = None

    def __init__(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        time1: Optional[int] = None,
        time2: Optional[int] = None,
    ):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.time1 = time1
        self.time2 = time2

    @property
    def time1(self):
        return self.__time1

    @time1.setter
    def time1(self, value: int):
        if self.time2 is not None and value is not None and value > self.time2:
            self.__time1 = self.time2
            self.__time2 = value
        else:
            self.__time1 = value

    @property
    def time2(self):
        return self.__time2

    @time2.setter
    def time2(self, value: int):
        if self.time1 is not None and value is not None and value < self.time1:
            self.__time2 = self.time1
            self.__time1 = value
        else:
            self.__time2 = value

    def __str__(self):
        if self.time1 is not None and self.time2 is not None:
            return f"\\{self.tag}({Format.format_float(self.x1)},{Format.format_float(self.y1)},{Format.format_float(self.x2)},{Format.format_float(self.y2)},{self.time1},{self.time2})"
        return f"\\{self.tag}({Format.format_float(self.x1)},{Format.format_float(self.y1)},{Format.format_float(self.x2)},{Format.format_float(self.y2)})"


@dataclass
class AssValidTagPosition(AssTagPosition):
    x: float
    y: float

    def __str__(self):
        return (
            f"\\{self.tag}({Format.format_float(self.x)},{Format.format_float(self.y)})"
        )



from ..ass_tag_wrap_style import AssTagWrapStyle, WrapStyle
from dataclasses import dataclass


@dataclass
class AssValidTagWrapStyle(AssTagWrapStyle):
    style: WrapStyle

    def __str__(self):
        return f"\\{self.tag}{self.style.value}"



from ..ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssInvalidTag(AssTag):
    """
    An AssInvalidTag is used when ass_tag_analyzer cannot know what is the value of an tag because it depend of the style of the line.
    Ex:
        "{\\i}Example" would return an AssInvalidTagItalic, because the text "Example" will totally depend on the style of the line.
        So, if the style is italic, then "Example" will be display in italic and vice-versa.

    Attributes:
        text (str): The text of an invalid tag. At 99% of the time, the text will be empty.
            Ex:
                "\\b99" would return AssInvalidTagBold("99") instanc which is a AssInvalidTag.
    """

    text: str

    def __str__(self):
        return f"\\{self.tag}{self.text}"



from ..ass_tag_color import (
    AssTagPrimaryColor,
    AssTagSecondaryColor,
    AssTagOutlineColor,
    AssTagBackgroundColor,
)
from dataclasses import dataclass


@dataclass
class AssValidTagPrimaryColor(AssTagPrimaryColor):
    red: int
    green: int
    blue: int

    def __str__(self):
        if self.is_short_tag:
            return f"\\{self.short_tag}&H{self.blue:02X}{self.green:02X}{self.red:02X}&"
        return f"\\{self.tag}&H{self.blue:02X}{self.green:02X}{self.red:02X}&"


@dataclass
class AssValidTagSecondaryColor(AssTagSecondaryColor):
    red: int
    green: int
    blue: int

    def __str__(self):
        return f"\\{self.tag}&H{self.blue:02X}{self.green:02X}{self.red:02X}&"


@dataclass
class AssValidTagOutlineColor(AssTagOutlineColor):
    red: int
    green: int
    blue: int

    def __str__(self):
        return f"\\{self.tag}&H{self.blue:02X}{self.green:02X}{self.red:02X}&"


@dataclass
class AssValidTagBackgroundColor(AssTagBackgroundColor):
    red: int
    green: int
    blue: int

    def __str__(self):
        return f"\\{self.tag}&H{self.blue:02X}{self.green:02X}{self.red:02X}&"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagBlurEdges(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L768-L780

    @property
    def tag(self) -> str:
        return "be"


@dataclass
class AssTagBlurEdgesGauss(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L381-L391

    @property
    def tag(self) -> str:
        return "blur"



from ..ass_tag_alpha import (
    AssTagAlpha,
    AssTagPrimaryAlpha,
    AssTagSecondaryAlpha,
    AssTagOutlineAlpha,
    AssTagBackgroundAlpha,
)
from dataclasses import dataclass


@dataclass
class AssValidTagAlpha(AssTagAlpha):
    value: int

    def __str__(self):
        return f"\\{self.tag}&H{self.value:02X}&"

@dataclass
class AssValidTagPrimaryAlpha(AssTagPrimaryAlpha):
    value: int

    def __str__(self):
        return f"\\{self.tag}&H{self.value:02X}&"


@dataclass
class AssValidTagSecondaryAlpha(AssTagSecondaryAlpha):
    value: int

    def __str__(self):
        return f"\\{self.tag}&H{self.value:02X}&"


@dataclass
class AssValidTagOutlineAlpha(AssTagOutlineAlpha):
    value: int

    def __str__(self):
        return f"\\{self.tag}&H{self.value:02X}&"


@dataclass
class AssValidTagBackgroundAlpha(AssTagBackgroundAlpha):
    value: int

    def __str__(self):
        return f"\\{self.tag}&H{self.value:02X}&"



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_font_scale import AssTagFontScale, AssTagFontXScale, AssTagFontYScale
from dataclasses import dataclass


@dataclass
class AssInvalidTagFontXScale(AssTagFontXScale, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagFontYScale(AssTagFontYScale, AssInvalidTag):
    pass



from .ass_item import (
    AssComment,
    AssItem,
    AssTagListOpening,
    AssTagListEnding,
    AssText,
    AssDraw,
)
from .ass_tag_alignment import Alignment, LegacyAlignment, AssTagAlignment
from .ass_tag_alpha import (
    AssTagAlpha,
    AssTagPrimaryAlpha,
    AssTagSecondaryAlpha,
    AssTagOutlineAlpha,
    AssTagBackgroundAlpha,
)
from .ass_tag_blur import AssTagBlurEdges, AssTagBlurEdgesGauss
from .ass_tag_border import AssTagBorder, AssTagXBorder, AssTagYBorder
from .ass_tag_clip import AssTagClipRectangle, AssTagClipVector
from .ass_tag_color import (
    AssTagPrimaryColor,
    AssTagSecondaryColor,
    AssTagOutlineColor,
    AssTagBackgroundColor,
)
from .ass_tag_fade import AssTagFade, AssTagFadeComplex
from .ass_tag_font_scale import AssTagFontScale, AssTagFontXScale, AssTagFontYScale
from .ass_tag_general import (
    AssTagAnimation,
    AssTagBaselineOffset,
    AssTagBold,
    AssTagDraw,
    AssTagFontEncoding,
    AssTagFontName,
    AssTagFontSize,
    AssTagItalic,
    AssTagLetterSpacing,
    AssTagResetStyle,
    AssTagRotationOrigin,
    AssTagStrikeout,
    AssTagUnderline,
)
from .ass_tag_karaoke import AssTagKaraoke, AssTagKaraokeFill, AssTagKaraokeOutline
from .ass_tag_position import AssTagMove, AssTagPosition
from .ass_tag_rotation import AssTagXRotation, AssTagYRotation, AssTagZRotation
from .ass_tag_shadow import AssTagShadow, AssTagXShadow, AssTagYShadow
from .ass_tag_shear import AssTagXShear, AssTagYShear
from .ass_tag_wrap_style import AssTagWrapStyle, WrapStyle

from .ass_valid_tag import *
from .ass_invalid_tag import *

__version__ = "0.0.1"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagPrimaryColor(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L705-L711
    is_short_tag: bool

    @property
    def tag(self) -> str:
        return "1c"

    @property
    def short_tag(self) -> str:
        return "c"


@dataclass
class AssTagSecondaryColor(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L712-L718

    @property
    def tag(self) -> str:
        return "2c"


@dataclass
class AssTagOutlineColor(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L719-L725

    @property
    def tag(self) -> str:
        return "3c"


@dataclass
class AssTagBackgroundColor(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L726-L732

    @property
    def tag(self) -> str:
        return "4c"



from .ass_item import AssTag
from dataclasses import dataclass
from enum import IntEnum


class WrapStyle(IntEnum):
    # https://aeg-dev.github.io/AegiSite/docs/3.2/ass_tags/#:~:text=Wrap%20style

    SMART_TOP = 0
    END_OF_LINE = 1
    NO_WORD = 2
    SMART_BOTTOM = 3


@dataclass
class AssTagWrapStyle(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L853-L857

    @property
    def tag(self) -> str:
        return "q"



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_border import AssTagBorder, AssTagXBorder, AssTagYBorder
from dataclasses import dataclass


@dataclass
class AssInvalidTagBorder(AssTagBorder, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagXBorder(AssTagXBorder, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagYBorder(AssTagYBorder, AssInvalidTag):
    pass



from ..ass_tag_alignment import AssTagAlignment, Alignment, LegacyAlignment
from dataclasses import dataclass
from typing import Union


@dataclass
class AssValidTagAlignment(AssTagAlignment):
    __alignment: Union[Alignment, LegacyAlignment]

    def __init__(
        self, alignment: Union[Alignment, LegacyAlignment], is_legacy_tag: bool = False
    ):
        self.is_legacy_tag = is_legacy_tag
        self.alignment = alignment

    def convert_to_a_tag(self):
        if not self.is_legacy_tag:
            self.is_legacy_tag = True
            self.alignment = LegacyAlignment[self.alignment.name]

    def convert_to_an_tag(self):
        if self.is_legacy_tag:
            self.is_legacy_tag = False
            self.alignment = Alignment[self.alignment.name]

    @property
    def alignment(self):
        return self.__alignment

    @alignment.setter
    def alignment(self, alignment: Union[Alignment, LegacyAlignment]):
        if self.is_legacy_tag:
            if isinstance(alignment, LegacyAlignment):
                self.__alignment = alignment
            else:
                raise ValueError("alignment need to be an LegacyAlignment instance")
        else:
            if isinstance(alignment, Alignment):
                self.__alignment = alignment
            else:
                raise ValueError("alignment need to be an Alignment instance")

    def __str__(self):
        if self.is_legacy_tag:
            return f"\\{self.legacy_tag}{self.alignment.value}"
        return f"\\{self.tag}{self.alignment.value}"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagShadow(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L817-L829

    @property
    def tag(self) -> str:
        return "shad"


@dataclass
class AssTagXShadow(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L329-L336

    @property
    def tag(self) -> str:
        return "xshad"


@dataclass
class AssTagYShadow(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L337-L344

    @property
    def tag(self) -> str:
        return "yshad"



from .ass_invalid_tag_alignment import AssInvalidTagAlignment
from .ass_invalid_tag_alpha import (
    AssInvalidTagAlpha,
    AssInvalidTagPrimaryAlpha,
    AssInvalidTagSecondaryAlpha,
    AssInvalidTagOutlineAlpha,
    AssInvalidTagBackgroundAlpha,
)
from .ass_invalid_tag_border import (
    AssInvalidTagBorder,
    AssInvalidTagXBorder,
    AssInvalidTagYBorder,
)
from .ass_invalid_tag_color import (
    AssInvalidTagPrimaryColor,
    AssInvalidTagSecondaryColor,
    AssInvalidTagOutlineColor,
    AssInvalidTagBackgroundColor,
)
from .ass_invalid_tag_font_scale import (
    AssInvalidTagFontXScale,
    AssInvalidTagFontYScale,
)
from .ass_invalid_tag_general import (
    AssInvalidTagBold,
    AssInvalidTagFontEncoding,
    AssInvalidTagFontName,
    AssInvalidTagFontSize,
    AssInvalidTagItalic,
    AssInvalidTagLetterSpacing,
    AssInvalidTagResetStyle,
    AssInvalidTagStrikeout,
    AssInvalidTagUnderline,
)
from .ass_invalid_tag_rotation import (
    AssInvalidTagZRotation,
)
from .ass_invalid_tag_shadow import (
    AssInvalidTagShadow,
    AssInvalidTagXShadow,
    AssInvalidTagYShadow,
)
from .ass_invalid_tag_wrap_style import AssInvalidTagWrapStyle



from abc import ABC, abstractmethod
from dataclasses import dataclass


class AssItem(ABC):
    @abstractmethod
    def __str__(self):
        pass


@dataclass
class AssTagListOpening(AssItem):
    def __str__(self):
        return "{"


@dataclass
class AssTagListEnding(AssItem):
    def __str__(self):
        return "}"


@dataclass
class AssText(AssItem):
    text: str

    def __str__(self):
        return self.text


@dataclass
class AssComment(AssItem):
    text: str

    def __str__(self):
        return self.text


@dataclass
class AssDraw(AssItem):
    text: str

    def __str__(self):
        return self.text


class AssTag(AssItem):
    @property
    @abstractmethod
    def tag(self) -> str:
        pass



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_wrap_style import AssTagWrapStyle
from dataclasses import dataclass


@dataclass
class AssInvalidTagWrapStyle(AssTagWrapStyle, AssInvalidTag):
    pass



import re


class TypeParser:
    # https://www.qnx.com/developers/docs/6.4.0/dinkum_en/c99/stdint.html#INT32_MIN
    INT32_MIN = -0x7FFFFFFF - 1
    # https://www.qnx.com/developers/docs/6.4.0/dinkum_en/c99/stdint.html#INT32_MAX
    INT32_MAX = 0x7FFFFFFF

    """
    From python documentation
        \s
            For Unicode (str) patterns:
                [...]. If the ASCII flag is used, only [ \t\n\r\f\v] is matched.
        \d
            For Unicode (str) patterns:
                [...]. If the ASCII flag is used only [0-9] is matched.
    Libass
        Space: https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_string.h#L27-L31
        Number: https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_string.h#L33-L36
    """
    FLOAT_STR_REGEX = re.compile(r"^\s*[+-]?\d*\.?\d*(?:[eE][+-]?\d+)?", flags=re.ASCII)
    INT_STR_REGEX = re.compile(r"^\s*[+-]?\d*", flags=re.ASCII)
    HEX_STR_REGEX = re.compile(r"^\s*[+-]?[\da-fA-F]*", flags=re.ASCII)

    @staticmethod
    def color_arg(color: int):
        """
        Parse text from an tag into an RGB color

        Parameters:

        Returns:
            The text parsed (red, green, blue)
        """

        red = color & 255
        green = (color >> 8) & 255
        blue = (color >> 16) & 255

        return red, green, blue

    @staticmethod
    def float_str_to_float(text: str) -> float:
        """
        Parse text from an tag into an float

        Parameters:
                text (str): Text. Ex: "+20.23"
        Returns:
            The float that represent the text

        Libass info:
            - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L46-L51
            - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_utils.h#L191-L196
        """
        match = TypeParser.FLOAT_STR_REGEX.search(text)

        try:
            return float(match.group(0))
        except (AttributeError, ValueError):
            return 0.0

    @staticmethod
    def int_str_to_int(text: str) -> int:
        """
        Parse text from an tag into an int

        Parameters:
                text (str): Text. Ex: "+20"
        Returns:
            The integer that represent the text

        Libass info:
            - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L39-L44
            - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_utils.h#L198-L204
        """
        match = TypeParser.INT_STR_REGEX.search(text)

        try:
            return TypeParser.int_to_int32(int(match.group(0)))
        except (AttributeError, ValueError):
            return 0

    @staticmethod
    def hex_str_to_int(text: str) -> int:
        """
        Parse text from an tag into an int

        Parameters:
                text (str): Text. Ex: "+20"
        Returns:
            The integer that represent the text

        Libass info:
            - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L39-L44
            - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_utils.h#L198-L204
        """
        match = TypeParser.HEX_STR_REGEX.search(text)

        try:
            return TypeParser.int_to_int32(int(match.group(0), 16))
        except (AttributeError, ValueError):
            return 0

    @staticmethod
    def int_to_int32(number: int) -> int:
        """
        Libass info:
            - https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_utils.h#L198-L204
        """
        return min(max(number, TypeParser.INT32_MIN), TypeParser.INT32_MAX)

    @staticmethod
    def strip_whitespace(text: str) -> str:
        """Remove whitespace from text.

        Inpired by: https://github.com/libass/libass/blob/5f57443f1784434fe8961275da08be6d6febc688/libass/ass_utils.c#L160-L174

        Parameters:
            text (str): An string.
        Returns:
            A string without whitespace at the beginning and at the end.
        """

        return text.strip(" \t")



from ..ass_tag_rotation import AssTagXRotation, AssTagYRotation, AssTagZRotation
from ass_tag_analyzer.ass_format import Format
from dataclasses import dataclass


@dataclass
class AssValidTagXRotation(AssTagXRotation):
    angle: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.angle)}"


@dataclass
class AssValidTagYRotation(AssTagYRotation):
    angle: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.angle)}"


@dataclass
class AssValidTagZRotation(AssTagZRotation):
    angle: float

    def __str__(self):
        if self.is_short_tag:
            return f"\\{self.short_tag}{Format.format_float(self.angle)}"
        return f"\\{self.tag}{Format.format_float(self.angle)}"



from ..ass_tag_clip import AssTagClipRectangle, AssTagClipVector
from dataclasses import dataclass


@dataclass
class AssValidTagClipRectangle(AssTagClipRectangle):
    x1: int
    y1: int
    x2: int
    y2: int

    def __str__(self):
        return f"\\{self.tag}({self.x1},{self.y1},{self.x2},{self.y2})"


@dataclass
class AssValidTagClipVector(AssTagClipVector):
    """
    If you wanna parse an Vector Clip, see the PyonFX library.
    Especially the Shape module: https://pyonfx.readthedocs.io/en/latest/reference/shape.html
    """

    path: str
    scale: int = 1

    def __str__(self):
        if self.scale == 1:
            return f"\\{self.tag}({self.path})"
        return f"\\{self.tag}({self.scale},{self.path})"



from .ass_item import AssTag
from dataclasses import dataclass
from enum import IntEnum


class Alignment(IntEnum):
    # https://aeg-dev.github.io/AegiSite/docs/3.2/ass_tags/#:~:text=Line%20alignment

    BOTTOM_LEFT = 1
    BOTTOM_CENTER = 2
    BOTTOM_RIGHT = 3
    MIDDLE_LEFT = 4
    MIDDLE_CENTER = 5
    MIDDLE_RIGHT = 6
    TOP_LEFT = 7
    TOP_CENTER = 8
    TOP_RIGHT = 9


class LegacyAlignment(IntEnum):
    # https://aeg-dev.github.io/AegiSite/docs/3.2/ass_tags/#:~:text=Line%20alignment%20(legacy)

    BOTTOM_LEFT = 1
    BOTTOM_CENTER = 2
    BOTTOM_RIGHT = 3
    MIDDLE_LEFT = 9
    MIDDLE_CENTER = 10
    MIDDLE_RIGHT = 11
    TOP_LEFT = 5
    TOP_CENTER = 6
    TOP_RIGHT = 7


@dataclass
class AssTagAlignment(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L540-L561
    is_legacy_tag: bool

    @property
    def tag(self) -> str:
        return "an"

    @property
    def legacy_tag(self) -> str:
        return "a"



from ass_tag_analyzer.ass_format import Format
from ..ass_tag_blur import AssTagBlurEdges, AssTagBlurEdgesGauss
from dataclasses import dataclass


@dataclass
class AssValidTagBlurEdges(AssTagBlurEdges):
    weight: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.weight)}"


@dataclass
class AssValidTagBlurEdgesGauss(AssTagBlurEdgesGauss):
    __weight: float

    def __init__(self, weight: float):
        self.weight = weight

    @property
    def weight(self):
        return self.__weight

    @weight.setter
    def weight(self, value: float):
        self.__weight = 0 if value < 0 else value

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.weight)}"



class Format:
    @staticmethod
    def _format_float(number: float) -> str:
        """
        Parameters:
            number (float): An float.
        Returns:
            The float converted to string.

            Ex:
                - 4.3004 --> "4.3"
                - 4.0054 --> "4.005"
                - 4.0056 --> "4.006"
        """
        return "{:.3f}".format(number).rstrip("0").rstrip(".")

    """
    All float tag like \\blur, \\be, \\fax, etc use this variable.
    If you wanna change how ass_tag_analyzer converted them into string, then you need to redefine format_float

    Ex:
        def my_new_way_to_format_float(number: float) -> str:
            return YOUR_WAY_TO_CONVERT_FLOAT_TO_STRING
        
        Format.format_float = my_new_way_to_format_float
    """
    format_float = _format_float



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_rotation import AssTagZRotation
from dataclasses import dataclass


@dataclass
class AssInvalidTagZRotation(AssTagZRotation, AssInvalidTag):
    pass



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagXRotation(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L487-L494

    @property
    def tag(self) -> str:
        return "frx"


@dataclass
class AssTagYRotation(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L495-L502

    @property
    def tag(self) -> str:
        return "fry"


@dataclass
class AssTagZRotation(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L503-L511
    is_short_tag: bool

    @property
    def tag(self) -> str:
        return "frz"

    @property
    def short_tag(self) -> str:
        return "fr"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagAlpha(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L523-L539

    @property
    def tag(self) -> str:
        return "alpha"


@dataclass
class AssTagPrimaryAlpha(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L733-L739

    @property
    def tag(self) -> str:
        return "1a"


@dataclass
class AssTagSecondaryAlpha(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L740-L746

    @property
    def tag(self) -> str:
        return "2a"


@dataclass
class AssTagOutlineAlpha(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L747-L753

    @property
    def tag(self) -> str:
        return "3a"


@dataclass
class AssTagBackgroundAlpha(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L754-L760

    @property
    def tag(self) -> str:
        return "4a"



from ass_tag_analyzer.ass_format import Format
from ..ass_tag_shear import AssTagXShear, AssTagYShear
from dataclasses import dataclass


@dataclass
class AssValidTagXShear(AssTagXShear):
    value: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.value)}"


@dataclass
class AssValidTagYShear(AssTagYShear):
    value: float

    def __str__(self):
        return f"\\{self.tag}{Format.format_float(self.value)}"



from .ass_item import AssTag
from dataclasses import dataclass


@dataclass
class AssTagXShear(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L345-L352

    @property
    def tag(self) -> str:
        return "fax"


@dataclass
class AssTagYShear(AssTag):
    # https://github.com/libass/libass/blob/44f6532daf5eb13cb1aa95f5449a77b5df1dd85b/libass/ass_parse.c#L353-L360

    @property
    def tag(self) -> str:
        return "fay"



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_alpha import (
    AssTagAlpha,
    AssTagPrimaryAlpha,
    AssTagSecondaryAlpha,
    AssTagOutlineAlpha,
    AssTagBackgroundAlpha,
)
from dataclasses import dataclass


@dataclass
class AssInvalidTagAlpha(AssTagAlpha, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagPrimaryAlpha(AssTagPrimaryAlpha, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagSecondaryAlpha(AssTagSecondaryAlpha, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagOutlineAlpha(AssTagOutlineAlpha, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagBackgroundAlpha(AssTagBackgroundAlpha, AssInvalidTag):
    pass



from ..ass_tag_fade import AssTagFade, AssTagFadeComplex
from dataclasses import dataclass


@dataclass
class AssValidTagFade(AssTagFade):
    time1: int
    time2: int

    def __str__(self):
        return f"\\{self.tag}({self.time1},{self.time2})"


@dataclass
class AssValidTagFadeComplex(AssTagFadeComplex):
    alpha1: int
    alpha2: int
    alpha3: int
    time1: int
    time2: int
    time3: int
    time4: int

    def __str__(self):
        return f"\\{self.tag}({self.alpha1},{self.alpha2},{self.alpha3},{self.time1},{self.time2},{self.time3},{self.time4})"



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_color import (
    AssTagPrimaryColor,
    AssTagSecondaryColor,
    AssTagOutlineColor,
    AssTagBackgroundColor,
)
from dataclasses import dataclass


@dataclass
class AssInvalidTagPrimaryColor(AssTagPrimaryColor, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagSecondaryColor(AssTagSecondaryColor, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagOutlineColor(AssTagOutlineColor, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagBackgroundColor(AssTagBackgroundColor, AssInvalidTag):
    pass



from .ass_invalid_tag import AssInvalidTag
from ..ass_tag_shadow import AssTagShadow, AssTagXShadow, AssTagYShadow
from dataclasses import dataclass


@dataclass
class AssInvalidTagShadow(AssTagShadow, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagXShadow(AssTagXShadow, AssInvalidTag):
    pass


@dataclass
class AssInvalidTagYShadow(AssTagYShadow, AssInvalidTag):
    pass



from ass_tag_analyzer.ass_type_parser import TypeParser


def test_color_arg():
    assert TypeParser.color_arg(int("FA9632", 16)) == (50, 150, 250)


def test_float_str_to_float():

    # Whitespace and invalid text
    assert TypeParser.float_str_to_float(" \t\v20.39445InvalidText") == 20.39445

    # Invalid text at the beginning
    assert TypeParser.float_str_to_float("InvalidText-20") == 0

    # Multiple dot
    assert TypeParser.float_str_to_float("20.39.445") == 20.39

    # One dot without number
    assert TypeParser.float_str_to_float("20.") == 20

    # Multiple +-
    assert TypeParser.float_str_to_float("+-20.39.445") == 0

    # Integer
    assert TypeParser.float_str_to_float("20") == 20

    # +-
    assert TypeParser.float_str_to_float("+20") == 20
    assert TypeParser.float_str_to_float("-20") == -20

    # Exponent
    assert TypeParser.float_str_to_float("2e+2") == 2 * pow(10, 2)
    assert TypeParser.float_str_to_float("2e-2") == 2 * pow(10, -2)
    assert TypeParser.float_str_to_float("2e2") == 2 * pow(10, 2)
    assert TypeParser.float_str_to_float("2e") == 2
    assert TypeParser.float_str_to_float("2.e10") == 2 * pow(10, 10)
    assert TypeParser.float_str_to_float("e10") == 0


def test_int_str_to_int():

    # Whitespace and invalid text
    assert TypeParser.int_str_to_int("\t\v20.39445InvalidText") == 20

    # Invalid text at the beginning
    assert TypeParser.int_str_to_int("InvalidText-20") == 0

    # Multiple dot
    assert TypeParser.int_str_to_int("20.39.445") == 20

    # Multiple +-
    assert TypeParser.int_str_to_int("+-20.39.445") == 0

    # Integer
    assert TypeParser.int_str_to_int("20") == 20

    # +-
    assert TypeParser.int_str_to_int("+20") == 20
    assert TypeParser.int_str_to_int("-20") == -20


def test_hex_str_to_int():

    # Whitespace and invalid text
    assert TypeParser.hex_str_to_int("\t\v6BInvalidText") == 107

    # Invalid text at the beginning
    assert TypeParser.hex_str_to_int("InvalidText-20") == 0

    # Multiple dot
    assert TypeParser.hex_str_to_int("6B.39.445") == 107

    # Multiple +-
    assert TypeParser.hex_str_to_int("+-6B.39.445") == 0

    # Integer
    assert TypeParser.hex_str_to_int("6B") == 107

    # +-
    assert TypeParser.hex_str_to_int("+6B") == 107
    assert TypeParser.hex_str_to_int("-6B") == -107

    # With lowercase characters
    assert TypeParser.hex_str_to_int("-6b") == -107
